from typing import List, Tuple
import pandas as pd
from talp_pages.io.run_folders import RunFolder
from talp_pages.io.dataframe_handling import get_dfs
from dataclasses import dataclass
from datetime import datetime


@dataclass
class TimeSeriesAnalysisArguments:
    run_folder: RunFolder


@dataclass
class TimeSeriesAnalysisResult:
    ressource_configuration: str
    dataframe: pd.DataFrame
    latest_change: Tuple[datetime, str]  # Timestamp and git commit if available


class TimeSeriesAnalysis:
    def __init__(self, arguments: TimeSeriesAnalysisArguments) -> None:
        if not arguments:
            raise ValueError("No arguments given to TimeSeriesAnalysis")
        self.run_folder = arguments.run_folder
        # get df from run_folders
        self.df = get_dfs([self.run_folder], as_list=False)  # we only have one

    def get_results(
        self, ressource_configurations: List[str]
    ) -> List[TimeSeriesAnalysisResult]:
        # Returns a result for every region
        results = []
        for ressource_configuration in ressource_configurations:
            dataframe = self.df[self.df["ressourceLabel"] == ressource_configuration]
            dataframe = dataframe.sort_values(
                by=["timestamp", "elapsedTime"], ascending=[False, False]
            )
            row_of_newest_timestamp = dataframe.loc[dataframe["timestamp"].idxmax()]
            date = pd.to_datetime(row_of_newest_timestamp["timestamp"])

            git_commit = str(row_of_newest_timestamp["gitCommitShort"])

            latest_change = (date, git_commit)

            results.append(
                TimeSeriesAnalysisResult(
                    ressource_configuration=ressource_configuration,
                    dataframe=dataframe,
                    latest_change=latest_change,
                )
            )

        return results



from typing import List, Union
import pandas as pd
from pathlib import Path
import json
from datetime import datetime
import numpy as np
from talp_pages.io.run_folders import RunFolder
from talp_pages.common import (
    TALP_JSON_TIMESTAMP_KEY,
    TALP_JSON_METADATA_KEY,
    ExecutionMode,
    TALP_VERSION_KEY,
    TALP_IMPLICIT_REGION_NAME,
    TALP_JSON_POP_METRICS_KEY,
    TALP_RESOURCES_KEY,
)

import logging


def detect_execution_mode(df) -> ExecutionMode:
    # First check if for all its 1
    total_threads = df["totalNumCpus"] - df["totalNumMpiRanks"]
    if df["totalNumCpus"].sum() == len(df["totalNumCpus"]):
        # Only 1 totalNumCpu per entry -> all entries are serial
        return ExecutionMode.SERIAL
    elif df["totalNumMpiRanks"].sum() == 0:
        # Not all runs are serial, but no MPI Ranks:
        return ExecutionMode.OPENMP
    elif total_threads.sum() == 0:
        # We have MPI Ranks, but no additional CPUs #TODO Think of LeWI
        return ExecutionMode.MPI
    else:
        return ExecutionMode.HYBRID


def __load_talp_json_3_5(json_input, json_path):
    # we do this a bit ugly for now and reorganize the json a bit:
    for region in json_input[TALP_JSON_POP_METRICS_KEY].keys():
        json_input[TALP_JSON_POP_METRICS_KEY][region]["regionName"] = region

    df = pd.DataFrame(json_input[TALP_JSON_POP_METRICS_KEY].values())
    try:
        df["runTimestamp"] = pd.to_datetime(json_input[TALP_JSON_TIMESTAMP_KEY])
    except KeyError:
        df["runTimestamp"] = pd.to_datetime(datetime.now())
        logging.critical(
            f"Could not read {TALP_JSON_TIMESTAMP_KEY} key in {json_path}. Using current time instead"
        )
    try:
        metadata = json_input[TALP_JSON_METADATA_KEY]
    except KeyError:
        metadata = {}
    df["metadata"] = json.dumps(metadata)

    if metadata:
        df["gitCommitShort"] = metadata["gitCommit"][:7]  # Git Short shas have 7 chars
        try:
            df["gitTimestamp"] = pd.to_datetime(metadata["gitTimestamp"])
            df["timestamp"] = pd.to_datetime(metadata["gitTimestamp"])
        except KeyError:
            df["timestamp"] = df["runTimestamp"]
    else:
        df["timestamp"] = df["runTimestamp"]
        df["gitCommitShort"] = ""

    try:
        resources = json_input[TALP_RESOURCES_KEY]
    except KeyError:
        resources = {"numCpus": 0, "numMpiRanks": 0}

    num_cpus = resources["numCpus"]
    num_mpi_ranks = resources["numMpiRanks"]
    df["totalNumCpus"] = num_cpus
    df["totalNumMpiRanks"] = num_mpi_ranks

    df["dlbVersion"] = "3.5"
    df["jsonFile"] = str(json_path)
    df["implicitRegionName"] = TALP_IMPLICIT_REGION_NAME

    # Compute some additional data

    invalid_cycles = df["cycles"] < 0
    invalid_instructions = df["instructions"] < 0
    # CLAMP cycles and instructions
    df.loc[invalid_cycles, "cycles"] = 0
    df.loc[invalid_instructions, "instructions"] = 0
    df["IPC"] = df["instructions"] / df["cycles"]
    df["IPC"] = df["IPC"].fillna(np.inf)

    if (sum(invalid_cycles) > 0) or sum(invalid_instructions) > 0:
        logging.warning(
            f"Invalid PAPI Counter values observed in {json_path}. Please check PAPI and DLB installations"
        )

    df["frequency"] = df["cycles"] / df["usefulTime"]
    # Normalize to seconds
    df["elapsedTime"] = df["elapsedTime"] * 1e-9

    execution_mode = detect_execution_mode(df)
    if execution_mode == ExecutionMode.HYBRID:
        # Then we do the X
        df["totalThreads"] = (
            df.loc[:, "totalNumCpus"] / df.loc[:, "totalNumMpiRanks"]
        ).astype(int)
        df["ressourceLabel"] = (
            df.loc[:, "totalNumMpiRanks"].astype(str)
            + "xMPI "
            + df.loc[:, "totalThreads"].astype(str)
            + "xOpenMP"
        )
    elif execution_mode == ExecutionMode.SERIAL:
        df["ressourceLabel"] = ExecutionMode.SERIAL.value
    else:
        df["ressourceLabel"] = (
            df.loc[:, "totalNumCpus"].astype(str) + f"x{execution_mode.value}"
        )

    df["executionMode"] = execution_mode.value
    # finally sort
    df = df.sort_values(by=["elapsedTime"], ascending=False)
    return df


def load_talp_json_df(json_path: Path) -> pd.DataFrame:
    with open(json_path) as file:
        run_json = json.load(file)
        df = None
        # First check version
        if run_json[TALP_VERSION_KEY].startswith("3.5"):
            df = __load_talp_json_3_5(run_json, json_path)
        else:
            raise ValueError(
                f"Unable to find parser for {run_json[TALP_VERSION_KEY]}. Please use a compatible TALP-Pages release for DLB version {run_json[TALP_VERSION_KEY]}"
            )
        return df


def get_dfs(
    run_folders: List[RunFolder], as_list=True
) -> Union[List[pd.DataFrame], pd.DataFrame]:
    dfs_outer = []
    for run_folder in run_folders:
        dfs_inner = []
        for json_entry in run_folder.jsons:
            df = load_talp_json_df(json_entry)
            df["runFolder"] = str(run_folder.relative_path)
            dfs_inner.append(df)

        # Now combine the inner semantically clustered jsons and add to list
        df_of_run_folder = pd.concat(dfs_inner)
        dfs_outer.append(df_of_run_folder)

    if as_list:
        return dfs_outer
    else:
        return pd.concat(dfs_outer)


def get_ressource_combinations(run_folder: RunFolder):
    df = get_dfs([run_folder], as_list=True)[0]  # we only have one
    return (
        df.sort_values(by="totalNumCpus")["ressourceLabel"].drop_duplicates().to_numpy()
    )



import numpy as np
import plotly.graph_objects as go
from plotly.colors import sample_colorscale

from talp_pages.common import TALP_METRIC_TO_NAME_MAP, TALP_METRIC_TO_NESTING_MAP
from talp_pages.common import ExecutionMode
from talp_pages.analysis.scaling_analysis import ScalingAnalysisResult


class ScalingTableHtmlPlot:
    def __init__(self, result: ScalingAnalysisResult):
        self.result = result

    def _get_table(self):
        ressource_combinations = self.result.ressource_combinations
        df_datapoints = self.result.dataframe_list

        if ExecutionMode.HYBRID.value in self.result.execution_modes:
            eff_metrics = [
                "globalEfficiency",
                "parallelEfficiency",
                "mpiParallelEfficiency",
                "mpiCommunicationEfficiency",
                "mpiLoadBalance",
                "mpiLoadBalanceIn",
                "mpiLoadBalanceOut",
                "ompParallelEfficiency",
                "ompSchedulingEfficiency",
                "ompLoadBalance",
                "ompSerializationEfficiency",
                "computationScalability",
                "instructionScaling",
                "ipcScaling",
                "frequencyScaling",
            ]
            abs_metrics = ["IPC", "frequency", "elapsedTime"]

        elif ExecutionMode.MPI.value in self.result.execution_modes:
            eff_metrics = [
                "globalEfficiency",
                "parallelEfficiency",
                "mpiParallelEfficiency",
                "mpiCommunicationEfficiency",
                "mpiLoadBalance",
                "mpiLoadBalanceIn",
                "mpiLoadBalanceOut",
                "computationScalability",
                "instructionScaling",
                "ipcScaling",
                "frequencyScaling",
            ]
            abs_metrics = ["IPC", "frequency", "elapsedTime"]

        elif ExecutionMode.OPENMP.value in self.result.execution_modes:
            eff_metrics = [
                "globalEfficiency",
                "parallelEfficiency",
                "ompParallelEfficiency",
                "ompSchedulingEfficiency",
                "ompLoadBalance",
                "ompSerializationEfficiency",
                "computationScalability",
                "instructionScaling",
                "ipcScaling",
                "frequencyScaling",
            ]
            abs_metrics = ["IPC", "frequency", "elapsedTime"]

        elif ExecutionMode.SERIAL.value in self.result.execution_modes:
            eff_metrics = [
                "globalEfficiency",
                "computationScalability",
                "instructionScaling",
                "ipcScaling",
                "frequencyScaling",
            ]
            abs_metrics = ["IPC", "frequency", "elapsedTime"]

        metrics = eff_metrics + abs_metrics

        header = ["Metrics"]
        header.extend(
            [f"<b>{combination}</b>" for combination in ressource_combinations]
        )
        aligns = ["left"]
        aligns.extend(["center" for _ in metrics])
        values = [
            [
                f"{'-'*TALP_METRIC_TO_NESTING_MAP[metric]} {TALP_METRIC_TO_NAME_MAP[metric]}"
                for metric in metrics
            ]
        ]
        numeric_values = [
            datapoint[metrics].to_numpy().squeeze().round(2)
            for datapoint in df_datapoints
        ]
        values.extend(numeric_values)

        colors = [np.full_like(metrics, "rgb(250, 250, 250)").squeeze()]
        colors.extend(
            [
                sample_colorscale(
                    "rdylgn", np.clip(numeric_value, 0, 1), low=0.0, high=1.0
                )
                for numeric_value in numeric_values
            ]
        )
        # reset color for absolute metrics
        for col in colors[1:]:
            col[len(eff_metrics) : len(metrics)] = [
                "rgb(250, 250, 250)" for _ in abs_metrics
            ]

        fig = go.Figure(
            data=[
                go.Table(
                    header=dict(
                        values=header,
                        # line_color='white', fill_color='white',
                        align=aligns,
                        font=dict(color="black", size=20),
                    ),
                    cells=dict(
                        values=values,
                        line_color=colors,
                        fill_color=colors,
                        height=30,
                        align=aligns,
                        font=dict(color="black", size=16),
                    ),
                )
            ]
        )
        return fig

    def get_html(self) -> str:
        fig = self._get_table()
        fig.update_layout(
            autosize=True,
            height=650,
            margin=dict(l=50, r=50, b=0, t=0, pad=4),
            font=dict(
                family="Roboto Condensed, monospace",
                # size=18,
            ),
        )
        return fig.to_html(full_html=False)



#!/usr/bin/env python

import argparse
import os
import logging
from pathlib import Path
from shutil import rmtree

from talp_pages.io.run_folders import get_run_folders
from talp_pages.common import TALP_IMPLICIT_REGION_NAME
from talp_pages.reports.ci_report import CIReport, CIReportArgs


def add_arguments(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-i",
        "--input-path",
        dest="path",
        help="Root path to search the *.json files in",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output-path",
        dest="output",
        help="relative path for the output",
        required=True,
    )
    parser.add_argument(
        "--regions",
        nargs="+",
        help="Names of regions of that will be highlited seperated by spaces",
        dest="regions",
    )
    parser.add_argument(
        "--region-for-badge",
        help=f"Name of region that will be used to generate the performance badges (Default: {TALP_IMPLICIT_REGION_NAME})",
        dest="region_for_badge",
        default=TALP_IMPLICIT_REGION_NAME,
    )
    parser.add_argument(
        "--badge",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Generate Parallel effiency badge .svg (Requires internet)",
    )
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        dest="log_level",
        help="Logger Level",
        default="WARNING",
    )

    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        dest="force",
        help="Overwrite the output directory if present",
    )


def _verify_input(args):
    path = None

    # Check if the JSON file exists
    if os.path.exists(args.path):
        found_a_json = False
        for _, _, filenames in os.walk(args.path):
            for _ in [f for f in filenames if f.endswith(".json")]:
                found_a_json = True

        if not found_a_json:
            logging.error(
                f"The specified path '{args.path}' doesnt contain any .json files "
            )
            raise Exception("Path empty of .json files")
    else:
        logging.error(f"The specified path '{args.path}' does not exist.")
        raise Exception("Not existing path")

    # Check if output is empty
    if os.path.exists(args.output):
        if len(os.listdir(args.output)) != 0:
            # non-empty output path
            if args.force:
                rmtree(args.output)
            else:
                logging.error(
                    f"The specified output folder '{args.output}' is not empty"
                )
                raise Exception("Non Empty output path")

    path = args.path
    output = args.output
    regions = [TALP_IMPLICIT_REGION_NAME]
    if args.regions:
        regions.extend(args.regions)

    region_for_badge = args.region_for_badge

    return path, output, regions, region_for_badge


def main(parsed_args):
    log_level = getattr(logging, parsed_args.log_level.upper(), None)
    logging.basicConfig(level=log_level)

    input_path, output_path, selected_regions, region_for_badge = _verify_input(
        parsed_args
    )

    run_folders = get_run_folders(input_path)
    if not run_folders:
        logging.fatal(
            f"Cannot find any valid folder structure containing jsons in {input_path}."
        )
        logging.fatal(
            "Please consult the documenation to learn what folder strucuture TALP-Pages expects."
        )
        raise ValueError("Empty run folders")
    print(
        f"Generating CI-Report for sub-folders {[str(run_folder.relative_path) for run_folder in run_folders]}"
    )
    report = CIReport(
        CIReportArgs(run_folders, selected_regions, region_for_badge, parsed_args.badge)
    )
    report.write_to_folder(Path(output_path))



#!/usr/bin/env python

import argparse

import talp_pages.cli.ci_report as ci_report
import talp_pages.cli.metadata as metadata
import talp_pages.cli.download_gitlab as download_gitlab
import talp_pages.cli.download_github as download_github
from talp_pages.common import TALP_PAGES_VERSION


def main():
    parser = argparse.ArgumentParser(
        prog="talp", description="TALP-Pages CLI interface."
    )

    parser.add_argument("--version", action="version", version=TALP_PAGES_VERSION)
    subparsers = parser.add_subparsers(title="features", help="", dest="features")

    ci_report_parser = subparsers.add_parser(
        "ci-report",
        help="Generate a HTML report containing performance evolution and insight plots.",
    )
    metadata_parser = subparsers.add_parser("metadata", help="Add metdata to JSONS")
    download_gitlab_parser = subparsers.add_parser(
        "download-gitlab", help="Download GitLab Artifacts"
    )
    download_github_parser = subparsers.add_parser(
        "download-github", help="Download GitHub Artifacts"
    )

    ci_report.add_arguments(ci_report_parser)
    metadata.add_arguments(metadata_parser)
    download_gitlab.add_arguments(download_gitlab_parser)
    download_github.add_arguments(download_github_parser)

    args = parser.parse_args()

    print(f"Welcome to TALP-Pages ({TALP_PAGES_VERSION})")

    if args.features == "ci-report":
        ci_report.main(args)
    elif args.features == "metadata":
        metadata.main(args)
    elif args.features == "download-gitlab":
        download_gitlab.main(args)
    elif args.features == "download-github":
        download_github.main(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from urllib.request import urlopen, Request
from talp_pages.analysis.badge_analysis import BadgeResult


class Badge:
    def __init__(self, result: BadgeResult) -> None:
        self.result = result
        rc = result.ressource_configuration.replace(" ", "_")
        self.file_name = f"badge_{rc}.svg"

    def get_content(self):
        if not self.result.parallel_eff:
            raise ValueError("Empty result passed to Badge")
        parallel_efficiency = round(self.result.parallel_eff, 2)
        if self.result.parallel_eff < 0.6:
            bagde_url = f"https://img.shields.io/badge/Parallel_efficiency-{parallel_efficiency}-red"
        elif self.result.parallel_eff < 0.8:
            bagde_url = f"https://img.shields.io/badge/Parallel_efficiency-{parallel_efficiency}-orange"
        else:
            bagde_url = f"https://img.shields.io/badge/Parallel_efficiency-{parallel_efficiency}-green"

        return (
            urlopen(Request(url=bagde_url, headers={"User-Agent": "Mozilla"}))
            .read()
            .decode("utf-8")
        )



from talp_pages.common import TALP_PAGES_VERSION
import pathlib
from jinja2 import Environment, FileSystemLoader
from datetime import datetime

TALP_PAGES_TEMPLATE_PATH = pathlib.Path(__file__).parent.joinpath("templates").resolve()


class HTMLSite:
    def __init__(self, template_name: str, file_name: str) -> None:
        self.page_metadata = {
            "page_metadata": {
                "timestamp_string": datetime.now().isoformat(timespec="seconds"),
                "version": TALP_PAGES_VERSION,
                "home": "../index.html",
            }
        }
        self.jinja_env = Environment(loader=FileSystemLoader(TALP_PAGES_TEMPLATE_PATH))
        self.template_name = template_name
        self.template = self.jinja_env.get_template(self.template_name)
        self.file_name = file_name

    def render_template(self, **context) -> str:
        # Render the template with the provided context
        return self.template.render({**context, **self.page_metadata})

    def get_content(self) -> str:
        raise NotImplementedError("Subclass needs to overrride this method")


class HTMLChildPage(HTMLSite):
    def __init__(self, template_name: str, file_name: str, icon: str) -> None:
        super().__init__(template_name, file_name)
        self.latest_change = None
        self.icon = icon



"""
File declaring some global scoped variables and functions
"""

from enum import Enum


class ExecutionMode(Enum):
    SERIAL = "Serial"
    OPENMP = "OpenMP"
    MPI = "MPI"
    HYBRID = "Hybrid MPI+OpenMP"


TALP_PAGES_VERSION = "3.5.1"
TALP_METRIC_TO_NAME_MAP = {
    "globalEfficiency": "Global Effiency",
    "globalEfficiencyBySpeedup": "Global Effiency from Speedup",
    "parallelEfficiency": "Parallel efficiency",
    "computationScalability": "Computation Scalability",
    "averageIPC": "Average IPC",
    "cycles": "Cycles",
    "instructions": "Useful Instructions",
    "IPC": "Useful IPC",
    "frequency": "Frequency [GHz]",
    "mpiParallelEfficiency": "MPI Parallel efficiency",
    "mpiCommunicationEfficiency": "MPI Communication efficiency",
    "mpiLoadBalance": "MPI Load balance",
    "mpiLoadBalanceIn": "MPI In-node load balance",
    "mpiLoadBalanceOut": "MPI Inter-node load balance",
    "ompParallelEfficiency": "OpenMP Parallel efficiency",
    "ompLoadBalance": "OpenMP Load balance",
    "ompSchedulingEfficiency": "OpenMP Scheduling efficiency",
    "ompSerializationEfficiency": "OpenMP Serialization efficiency",
    "instructionScaling": "Instructions scaling",
    "ipcScaling": "IPC scaling",
    "frequencyScaling": "Frequency scaling",
    "elapsedTime": "Elapsed time [s]",
    "speedup": "Speedup",
}

TALP_METRIC_TO_NESTING_MAP = {
    "globalEfficiency": 0,
    "globalEfficiencyBySpeedup": 0,
    "parallelEfficiency": 1,
    "computationScalability": 1,
    "averageIPC": 0,
    "cycles": 0,
    "instructions": 0,
    "IPC": 0,
    "frequency": 0,
    "mpiParallelEfficiency": 2,
    "mpiCommunicationEfficiency": 3,
    "mpiLoadBalance": 3,
    "mpiLoadBalanceIn": 4,
    "mpiLoadBalanceOut": 4,
    "ompParallelEfficiency": 2,
    "ompLoadBalance": 3,
    "ompSchedulingEfficiency": 3,
    "ompSerializationEfficiency": 3,
    "instructionScaling": 2,
    "ipcScaling": 2,
    "frequencyScaling": 2,
    "elapsedTime": 0,
    "speedup": 0,
}
# Some keys
TALP_IMPLICIT_REGION_NAME = "Global"
TALP_JSON_TIMESTAMP_KEY = "timestamp"
TALP_JSON_METADATA_KEY = "metadata"
TALP_VERSION_KEY = "dlbVersion"
TALP_JSON_POP_METRICS_KEY = "Application"
TALP_RESOURCES_KEY = "resources"



from plotly.subplots import make_subplots
import plotly.graph_objects as go
from typing import List
import numpy as np
from talp_pages.io.dataframe_handling import ExecutionMode
from talp_pages.analysis.time_series_analysis import TimeSeriesAnalysisResult

from talp_pages.common import TALP_IMPLICIT_REGION_NAME, TALP_METRIC_TO_NAME_MAP


class TimeSeriesHtmlPlot:
    def __init__(self, result: TimeSeriesAnalysisResult, regions: List[str]):
        self.result = result
        self.df = result.dataframe
        self.highlighted_regions = regions

    def __get_colors(self, regions: List[str]):
        region_to_colors = {}
        for region in regions:
            # compute average % of region
            mean_region = self.df.loc[
                self.df["regionName"] == region, "elapsedTime"
            ].mean()
            mean_implicit = self.df.loc[
                self.df["regionName"] == TALP_IMPLICIT_REGION_NAME, "elapsedTime"
            ].mean()
            percentage = mean_region / mean_implicit
            alpha = np.single(np.sqrt(percentage) + 0.2)
            alpha = np.clip(alpha, 0, 1)
            region_to_colors[region] = (
                f"rgba({np.random.randint(0,255)}, {np.random.randint(0,255)}, {np.random.randint(0,255)}, {alpha.round(2)})"
            )
        return region_to_colors

    def __get_scatter(self, region, metric, color, show_legend=False):
        if region in self.highlighted_regions:
            is_visible = True
        else:
            is_visible = "legendonly"

        return go.Scatter(
            x=self.df[self.df["regionName"] == region]["timestamp"],
            y=self.df[self.df["regionName"] == region][metric],
            text=self.df[self.df["regionName"] == region]["metadata"],
            mode="lines+markers",
            name=region,
            legendgroup=region,
            showlegend=show_legend,
            marker_color=color,
            line=dict(color=color),
            visible=is_visible,
            # hovertemplate='%{text}'),
        )

    def _get_figure(self):
        regions = self.df["regionName"].unique().tolist()
        if ExecutionMode.HYBRID.value in self.df["executionMode"].to_numpy():
            metrics = [
                "elapsedTime",
                "IPC",
                "instructions",
                "frequency",
                "parallelEfficiency",
                "mpiParallelEfficiency",
                "mpiCommunicationEfficiency",
                "mpiLoadBalance",
                "ompParallelEfficiency",
                "ompSchedulingEfficiency",
                "ompLoadBalance",
                "ompSerializationEfficiency",
            ]
            specs = [
                [{"colspan": 6}, None, None, None, None, None],  # elapsedTime
                [
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                ],  # 'IPC','instructions','frequency'
                [{"colspan": 6}, None, None, None, None, None],  # parallelEfficiency
                [{"colspan": 6}, None, None, None, None, None],  # mpiParallelEfficiency
                [
                    {"colspan": 3},
                    None,
                    None,
                    {"colspan": 3},
                    None,
                    None,
                ],  # 'mpiCommunicationEfficiency','mpiCommunicationEfficiency'
                [{"colspan": 6}, None, None, None, None, None],  # ompParallelEfficiency
                [
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                ],  # 'ompSchedulingEfficiency','ompLoadBalance','ompSerializationEfficiency'
            ]
            execution_mode = ExecutionMode.HYBRID

        elif ExecutionMode.MPI.value in self.df["executionMode"].to_numpy():
            metrics = [
                "elapsedTime",
                "IPC",
                "instructions",
                "frequency",
                "parallelEfficiency",
                "mpiParallelEfficiency",
                "mpiCommunicationEfficiency",
                "mpiLoadBalance",
            ]
            specs = [
                [{"colspan": 6}, None, None, None, None, None],  # elapsedTime
                [
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                ],  # 'IPC','instructions','frequency'
                [{"colspan": 6}, None, None, None, None, None],  # parallelEfficiency
                [{"colspan": 6}, None, None, None, None, None],  # mpiParallelEfficiency
                [
                    {"colspan": 3},
                    None,
                    None,
                    {"colspan": 3},
                    None,
                    None,
                ],  # 'mpiCommunicationEfficiency','mpiCommunicationEfficiency'
            ]
            execution_mode = ExecutionMode.MPI

        elif ExecutionMode.OPENMP.value in self.df["executionMode"].to_numpy():
            metrics = [
                "elapsedTime",
                "IPC",
                "instructions",
                "frequency",
                "parallelEfficiency",
                "ompParallelEfficiency",
                "ompSchedulingEfficiency",
                "ompLoadBalance",
                "ompSerializationEfficiency",
            ]
            specs = [
                [{"colspan": 6}, None, None, None, None, None],  # elapsedTime
                [
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                ],  # 'IPC','instructions','frequency'
                [{"colspan": 6}, None, None, None, None, None],  # parallelEfficiency
                [{"colspan": 6}, None, None, None, None, None],  # ompParallelEfficiency
                [
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                ],  # 'ompSchedulingEfficiency','ompLoadBalance','ompSerializationEfficiency'
            ]
            execution_mode = ExecutionMode.OPENMP

        else:
            # Serial
            metrics = [
                "elapsedTime",
                "IPC",
                "instructions",
                "frequency",
            ]
            specs = [
                [{"colspan": 6}, None, None, None, None, None],  # elapsedTime
                [
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                    {"colspan": 2},
                    None,
                ],  # 'IPC','instructions','frequency'
            ]
            execution_mode = ExecutionMode.SERIAL

        region_to_colors = self.__get_colors(regions)
        # sampled_colors = list(sample_colorscale(colorscale, len(regions)))

        # Create Big figure
        fig = make_subplots(
            rows=len(specs),
            cols=6,
            specs=specs,
            horizontal_spacing=0.07,
            vertical_spacing=0.05,
            subplot_titles=[TALP_METRIC_TO_NAME_MAP[metric] for metric in metrics],
            print_grid=False,
        )

        # Elapsed time
        metric = "elapsedTime"
        plots_done = 1
        for region in regions:
            fig.add_trace(
                self.__get_scatter(region, metric, region_to_colors[region], True),
                col=1,
                row=1,
            )
        fig["layout"]["yaxis"]["title"] = "Time in [s]"

        # IPC
        metric = "IPC"
        plots_done += 1
        for region in regions:
            fig.add_trace(
                self.__get_scatter(region, metric, region_to_colors[region]),
                col=1,
                row=2,
            )
        fig["layout"][f"yaxis{plots_done}"]["title"] = "IPC"

        # instructions
        metric = "instructions"
        plots_done += 1
        for region in regions:
            fig.add_trace(
                self.__get_scatter(region, metric, region_to_colors[region]),
                col=3,
                row=2,
            )
        fig["layout"][f"yaxis{plots_done}"]["title"] = "# instructions"

        # frequency
        metric = "frequency"
        plots_done += 1
        for region in regions:
            fig.add_trace(
                self.__get_scatter(region, metric, region_to_colors[region]),
                col=5,
                row=2,
            )
        fig["layout"][f"yaxis{plots_done}"]["title"] = "GHz"

        if (
            execution_mode == ExecutionMode.HYBRID
            or execution_mode == ExecutionMode.MPI
            or execution_mode == ExecutionMode.OPENMP
        ):
            # Parallel Effiency
            metric = "parallelEfficiency"
            plots_done += 1
            for region in regions:
                fig.add_trace(
                    self.__get_scatter(region, metric, region_to_colors[region]),
                    col=1,
                    row=3,
                )
            fig["layout"][f"yaxis{plots_done}"]["title"] = "Efficiency [0-1]"
            fig.update_layout(**{f"yaxis{plots_done}": dict(range=[0, 1])})

        if (
            execution_mode == ExecutionMode.HYBRID
            or execution_mode == ExecutionMode.MPI
        ):
            # communicationEfficiency
            metric = "mpiParallelEfficiency"
            plots_done += 1
            for region in regions:
                fig.add_trace(
                    self.__get_scatter(region, metric, region_to_colors[region]),
                    col=1,
                    row=4,
                )
            fig["layout"][f"yaxis{plots_done}"]["title"] = "Efficiency [0-1]"
            fig.update_layout(**{f"yaxis{plots_done}": dict(range=[0, 1])})

            metric = "mpiCommunicationEfficiency"
            plots_done += 1
            for region in regions:
                fig.add_trace(
                    self.__get_scatter(region, metric, region_to_colors[region]),
                    col=1,
                    row=5,
                )
            fig["layout"][f"yaxis{plots_done}"]["title"] = "Efficiency [0-1]"
            fig.update_layout(**{f"yaxis{plots_done}": dict(range=[0, 1])})

            metric = "mpiLoadBalance"
            plots_done += 1
            for region in regions:
                fig.add_trace(
                    self.__get_scatter(region, metric, region_to_colors[region]),
                    col=4,
                    row=5,
                )
            fig["layout"][f"yaxis{plots_done}"]["title"] = "Efficiency [0-1]"
            fig.update_layout(**{f"yaxis{plots_done}": dict(range=[0, 1])})

        if (
            execution_mode == ExecutionMode.OPENMP
            or execution_mode == ExecutionMode.HYBRID
        ):
            #'ompSchedulingEfficiency','ompLoadBalance','ompSerializationEfficiency',
            metric = "ompParallelEfficiency"
            plots_done += 1
            for region in regions:
                fig.add_trace(
                    self.__get_scatter(region, metric, region_to_colors[region]),
                    col=1,
                    row=6 if execution_mode == ExecutionMode.HYBRID else 4,
                )
            fig["layout"][f"yaxis{plots_done}"]["title"] = "Efficiency [0-1]"
            fig.update_layout(**{f"yaxis{plots_done}": dict(range=[0, 1])})

            metric = "ompSchedulingEfficiency"
            plots_done += 1
            for region in regions:
                fig.add_trace(
                    self.__get_scatter(region, metric, region_to_colors[region]),
                    col=1,
                    row=7 if execution_mode == ExecutionMode.HYBRID else 5,
                )
            fig["layout"][f"yaxis{plots_done}"]["title"] = "Efficiency [0-1]"
            fig.update_layout(**{f"yaxis{plots_done}": dict(range=[0, 1])})

            metric = "ompLoadBalance"
            plots_done += 1
            for region in regions:
                fig.add_trace(
                    self.__get_scatter(region, metric, region_to_colors[region]),
                    col=3,
                    row=7 if execution_mode == ExecutionMode.HYBRID else 5,
                )
            fig["layout"][f"yaxis{plots_done}"]["title"] = "Efficiency [0-1]"
            fig.update_layout(**{f"yaxis{plots_done}": dict(range=[0, 1])})

            metric = "ompSerializationEfficiency"
            plots_done += 1
            for region in regions:
                fig.add_trace(
                    self.__get_scatter(region, metric, region_to_colors[region]),
                    col=5,
                    row=7 if execution_mode == ExecutionMode.HYBRID else 5,
                )
            fig["layout"][f"yaxis{plots_done}"]["title"] = "Efficiency [0-1]"
            fig.update_layout(**{f"yaxis{plots_done}": dict(range=[0, 1])})

        fig.update_traces(mode="markers+lines")
        return fig

    def get_html(self) -> str:
        fig = self._get_figure()
        fig.update_layout(
            title="Performance metrics evolution",
            legend_title="Regions",
            autosize=True,
            height=2000,
            margin=dict(l=50, r=50, b=100, t=100, pad=4),
            font=dict(
                family="Roboto Condensed, monospace",
                # size=18,
            ),
        )
        return fig.to_html(full_html=False)



from talp_pages.html.site import HTMLChildPage
from talp_pages.analysis.time_series_analysis import TimeSeriesAnalysisResult
from typing import List
from talp_pages.plots.time_series_html_plot import TimeSeriesHtmlPlot


class TimeSeriesSite(HTMLChildPage):
    def __init__(self, result: TimeSeriesAnalysisResult, regions: List[str]) -> None:
        rc = result.ressource_configuration.replace(" ", "_")
        super().__init__(
            "time_series.jinja", file_name=f"evolution_{rc}.html", icon="bi-graph-up"
        )
        self.latest_change = result.latest_change
        self.result = result
        self.highlited_regions = regions

    def get_content(self):
        if not self.result:
            raise ValueError("Handed no results to TimeSeriesSites")
        plot = TimeSeriesHtmlPlot(result=self.result, regions=self.highlited_regions)
        html = self.render_template(figure=plot.get_html())
        return html



from talp_pages.html.site import HTMLSite


class CIReportIndexPage(HTMLSite):
    def __init__(self, subsites) -> None:
        super().__init__("index.jinja", "index.html")
        self.sites = subsites

    def get_content(self):
        html = self.render_template(sites_dict=self.sites, zip=zip)
        return html



#!/usr/bin/env python

import argparse
import os
import logging
from dataclasses import dataclass
from typing import Union
from talp_pages.io.run_folders import get_run_folders
import json
from datetime import datetime
import subprocess


def add_arguments(parser: argparse.ArgumentParser):
    # Adding argument for JSON file
    parser.add_argument(
        "-i",
        "--input-path",
        dest="path",
        help="Root path to search the *.json files in",
        required=True,
    )
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        dest="log_level",
        help="Logger Level",
        default="WARNING",
    )
    parser.add_argument(
        "--git-commit", dest="git_commit", help="Git commit to put into the metadata"
    )
    parser.add_argument(
        "--git-branch", dest="git_branch", help="Git branch to put into the metadata"
    )
    parser.add_argument(
        "--git-timestamp",
        dest="git_timestamp",
        help="Timestamp (in ISO 8601 format) of git commit to put into the metadata",
    )

    parser.add_argument(
        "--is-merge-request",
        action=argparse.BooleanOptionalAction,
        dest="is_merge_request",
        help="Git branch to put into the metadata",
    )


def _verify_input(args):
    path = None

    # Check if the JSON file existss
    if os.path.exists(args.path):
        found_a_json = False
        for _, _, filenames in os.walk(args.path):
            for _ in [f for f in filenames if f.endswith(".json")]:
                found_a_json = True

        if not found_a_json:
            logging.error(
                f"The specified path '{args.path}' doesnt contain any .json files "
            )
            raise Exception("Path empty of .json files")
    else:
        logging.error(f"The specified path '{args.path}' does not exist.")
        raise Exception("Not existing path")

    path = args.path

    return path


@dataclass
class Metadata:
    git_commit: str
    git_branch: str
    git_timestamp: datetime
    default_git_branch: str
    is_merge_request: bool

    def dict(self) -> dict:
        return {
            "gitCommit": self.git_commit,
            "gitBranch": self.git_branch,
            "gitTimestamp": self.git_timestamp.replace(tzinfo=None).isoformat()
            if self.git_timestamp
            else None,
            "defaultGitBranch": self.default_git_branch,
            "isMergeRequest": self.is_merge_request,
        }


UNKNOWN_METADATA = Metadata(
    "unkown commit", "unkown branch", None, "unkown default branch", False
)


def parse_date(date_string: str) -> datetime:
    # workaround until python 3.11 where datetime support reading timeszoned strings
    if date_string.endswith("Z"):
        date_string = date_string[:-1]
    dt = datetime.fromisoformat(date_string)
    return dt


def get_metadata(args: Union[argparse.Namespace, None]) -> Metadata:
    metadata = UNKNOWN_METADATA
    set_from_cli = False

    if args:
        # first try if its specified in the args, then dont use autodetect
        if args.git_commit:
            metadata.git_commit = f"{args.git_commit}"
            set_from_cli = True

        if args.git_branch:
            metadata.git_branch = args.git_branch
            set_from_cli = True

        if args.is_merge_request:
            metadata.is_merge_request = args.is_merge_request
            set_from_cli = True

        if args.git_timestamp:
            metadata.git_timestamp = parse_date(args.git_timestamp)

    if set_from_cli:
        logging.info(
            "User specified command line arguments, so we dont autodetect the metadata"
        )
        return metadata
    # gitlab Merge request
    if "CI_MERGE_REQUEST_ID" in os.environ:
        metadata.git_commit = os.environ.get(
            "CI_COMMIT_SHA", UNKNOWN_METADATA.git_commit
        )
        metadata.git_branch = os.environ.get(
            "CI_MERGE_REQUEST_SOURCE_BRANCH_NAME", UNKNOWN_METADATA.git_branch
        )
        metadata.git_timestamp = parse_date(
            os.environ.get("CI_COMMIT_TIMESTAMP", UNKNOWN_METADATA.git_timestamp)
        )
        metadata.default_git_branch = os.environ.get(
            "CI_DEFAULT_BRANCH", UNKNOWN_METADATA.default_git_branch
        )
        metadata.is_merge_request = True
        logging.info("Auto detected Metadata from GITLAB Merge Request")
        return metadata

    # gitlab normal pipeline
    if "CI_COMMIT_SHA" in os.environ:
        metadata.git_commit = os.environ.get(
            "CI_COMMIT_SHA", UNKNOWN_METADATA.git_commit
        )
        metadata.git_branch = os.environ.get(
            "CI_COMMIT_BRANCH", UNKNOWN_METADATA.git_branch
        )
        # Because github does not export the same information we have to do something slighly ugly
        timestamp_result = subprocess.run(
            ["git", "show", "-s", "--format=%cI", metadata.git_commit],
            capture_output=True,
            text=True,
        )
        metadata.git_timestamp = parse_date(timestamp_result.stdout.strip())
        metadata.is_merge_request = False
        logging.info("Auto detected Metadata from GITLAB normal pipeline")
        return metadata

    # Github Pull request
    if "GITHUB_HEAD_REF" in os.environ:
        metadata.git_commit = os.environ.get("GITHUB_SHA", UNKNOWN_METADATA.git_commit)
        metadata.git_branch = os.environ.get(
            "GITHUB_HEAD_REF", UNKNOWN_METADATA.git_branch
        )
        metadata.default_git_branch = "GitHub doenst expose the default branch via ENV"
        metadata.is_merge_request = True
        logging.info("Auto detected Metadata from Github Pull Request")
        return metadata

    # Github Normal action
    if "GITHUB_REF_NAME" in os.environ:
        metadata.git_commit = os.environ.get("GITHUB_SHA", UNKNOWN_METADATA.git_commit)
        metadata.git_branch = os.environ.get(
            "GITHUB_REF_NAME", UNKNOWN_METADATA.git_branch
        )
        metadata.default_git_branch = "GitHub doenst expose the default branch via ENV"
        metadata.is_merge_request = False
        logging.info("Auto detected Metadata from Github normal Action")
        return metadata

    logging.error(
        "Couldnt determine the metadata. Please use the command line interface to specify them."
    )
    return metadata


def add_metadata(metadata: Metadata, talp_json) -> None:
    metadata_dict = metadata.dict()

    if "metadata" in talp_json.keys():
        for key in metadata_dict.keys():
            if key in talp_json["metadata"].keys():
                logging.error(f"metadata.{key} already present in JSON")
                raise Exception(f"metadata.{key} already present in JSON")

        logging.info("Adding new metadata to existing metadata")
        already_present_metadata = talp_json["metadata"]
        talp_json["metadata"] = {**already_present_metadata, **metadata_dict}
    else:
        talp_json["metadata"] = metadata_dict


def main(parsed_args):
    path = _verify_input(parsed_args)

    log_level = getattr(logging, parsed_args.log_level.upper(), None)
    logging.basicConfig(level=log_level)

    # First detect the folder structure and append to the runs
    run_folders = get_run_folders(path)

    metadata = get_metadata(parsed_args)

    for run_folder in run_folders:
        for json_file_path in run_folder.jsons:
            # Open all the jsons but do in-place replacement of content
            talp_json = None
            with open(json_file_path, "r") as file:
                logging.debug(f"Start processing file {json_file_path}")
                talp_json = json.load(file)
                add_metadata(metadata, talp_json)
            with open(json_file_path, "w") as file:
                file.write(json.dumps(talp_json, indent=4))
                logging.debug(f"Done processing file {json_file_path}")



from pathlib import Path
import os

import logging

from talp_pages.html.site import HTMLSite
from talp_pages.svg.badge import Badge


def write_badge(output_path: Path, badge: Badge):
    os.makedirs(output_path, exist_ok=True)
    file_path = os.path.join(output_path, badge.file_name)
    logging.debug(f"Writing Badge {file_path}")
    with open(file_path, "w") as file:
        file.write(badge.get_content())


def write_site(output_path: Path, site: HTMLSite):
    os.makedirs(output_path, exist_ok=True)
    file_path = os.path.join(output_path, site.file_name)
    logging.debug(f"Writing Site {file_path}")
    with open(file_path, "w") as file:
        file.write(site.get_content())



from typing import List, Tuple
import pandas as pd
import numpy as np
import logging
from enum import Enum
from talp_pages.io.run_folders import RunFolder
from talp_pages.io.dataframe_handling import ExecutionMode
from talp_pages.io.dataframe_handling import get_dfs
from dataclasses import dataclass
from datetime import datetime

# Slighlty hacky workaround, but works for now to disable ugly warning
pd.options.mode.chained_assignment = None  # default='warn'


@dataclass
class ScalingAnalysisArguments:
    run_folder: RunFolder


class ScalingMode(Enum):
    WEAK = "Weak Scaling"
    STRONG = "Strong Scaling"


@dataclass
class ScalingAnalysisResult:
    region: str
    scaling_mode: ScalingMode
    scaling_table: pd.DataFrame
    execution_modes: ExecutionMode
    latest_change: Tuple[datetime, str]  # Timestamp and git commit if available
    ressource_combinations: List[str]
    dataframe_list: List[pd.DataFrame]


class ScalingAnalysis:
    def __init__(self, arguments: ScalingAnalysisArguments) -> None:
        if not arguments:
            raise ValueError("No ScalingAnalysisArguments given")
        self.run_folder = arguments.run_folder
        # get df from run_folders
        self.df = get_dfs([self.run_folder], as_list=False)

    def __detect_scaling_mode(self, dfs: List[pd.DataFrame]) -> ScalingMode:
        # Assume a ordered list where the ressource with the lowest is the first
        # Use instructions to estimate the weak or strong scaling
        # Assumption: IFF Weak scaling the number of instructions/totalNumCpus should be roughly the same within a bound of +- 50%
        # Default mode. WEAK
        if not dfs:
            return ScalingMode.WEAK

        base_case = dfs[0]

        if base_case["instructions"].values[0] <= 0:
            logging.warning(
                f"Unable to detect ScalingMode, as no instructions are available. Run with PAPI. Assuming {ScalingMode.WEAK}"
            )
            return ScalingMode.WEAK

        # Assume Weak scaling -> if condition violated to strong scaling
        for df in dfs:
            instruction_growth = (
                df["instructions"].to_numpy() / base_case["instructions"].to_numpy()
            )
            cpu_growth = df["numCpus"].to_numpy() / base_case["numCpus"].to_numpy()

            if (instruction_growth / cpu_growth) < 0.75:
                logging.info(
                    f"ScalingMode found to be Strong as {instruction_growth / cpu_growth} < 0.75"
                )
                return ScalingMode.STRONG

        return ScalingMode.WEAK

    def __compute_scaling_metrics(
        self, region: str
    ) -> Tuple[pd.DataFrame, pd.DataFrame, List[pd.DataFrame]]:
        # Computes the scaling metrics and returns a tuple with the first beeing just the basic_anylsis like table
        # and the second one being a ordered list of the dataframes making up the BA table.

        clean_df = self.df.loc[self.df["regionName"] == region]
        if clean_df.empty:
            raise ValueError("Region not available")

        ressource_combinations = (
            clean_df.sort_values(by="totalNumCpus")["ressourceLabel"]
            .drop_duplicates()
            .to_numpy()
        )
        df_datapoints = []
        for combination in ressource_combinations:
            datapoint = (
                clean_df.loc[clean_df["ressourceLabel"] == combination]
                .sort_values(by=["timestamp"], ascending=False)
                .head(1)
            )
            df_datapoints.append(datapoint)
        if not df_datapoints:
            logging.error(
                f"Not able to compute the scaling metrics for dataframe: {clean_df}"
            )
            # return (pd.DataFrame(),[])

        scaling_mode = self.__detect_scaling_mode(df_datapoints)

        clean_df.loc[:, "scalingMode"] = scaling_mode.value

        reference_case = df_datapoints[0]

        if scaling_mode == ScalingMode.STRONG:
            # Assume instructions stay constant
            clean_df["instructionScaling"] = (
                reference_case["instructions"] / clean_df.loc[:, "instructions"]
            )
        if scaling_mode == ScalingMode.WEAK:
            clean_df["instructionScaling"] = (
                reference_case["instructions"] / reference_case["numCpus"]
            ) * (1 / (clean_df["instructions"] / clean_df["numCpus"]))

        clean_df.loc[:, "ipcScaling"] = clean_df.loc[:, "IPC"] / reference_case["IPC"]
        clean_df.loc[:, "frequencyScaling"] = (
            clean_df.loc[:, "frequency"] / reference_case["frequency"]
        )

        # NO Papi cases
        clean_df.loc[clean_df["instructionScaling"] == 0, "instructionScaling"] = np.nan
        clean_df.loc[clean_df["ipcScaling"] == 0, "ipcScaling"] = np.nan
        clean_df["instructionScaling"] = clean_df["instructionScaling"].fillna(np.inf)
        clean_df["ipcScaling"] = clean_df["ipcScaling"].fillna(np.inf)
        clean_df["frequencyScaling"] = clean_df["frequencyScaling"].fillna(np.inf)

        clean_df.loc[:, "computationScalability"] = (
            clean_df["instructionScaling"]
            * clean_df["ipcScaling"]
            * clean_df["frequencyScaling"]
        )
        clean_df.loc[:, "globalEfficiency"] = (
            clean_df.loc[:, "parallelEfficiency"]
            * clean_df.loc[:, "computationScalability"]
        )
        clean_df.loc[:, "speedup"] = (
            reference_case["elapsedTime"] / clean_df["elapsedTime"]
        )
        clean_df.loc[:, "globalEfficiencyBySpeedup"] = clean_df.loc[:, "speedup"] / (
            clean_df.loc[:, "numCpus"] / reference_case["numCpus"]
        )

        df_datapoints = []
        for combination in ressource_combinations:
            datapoint = (
                clean_df[clean_df["ressourceLabel"] == combination]
                .sort_values(by=["timestamp"], ascending=False)
                .head(1)
                .fillna(np.inf)
            )
            df_datapoints.append(datapoint)
        return (pd.DataFrame(), df_datapoints, scaling_mode, ressource_combinations)

    def get_results(self, regions: List[str]) -> List[ScalingAnalysisResult]:
        # Returns a result for every region
        results = []
        for region in regions:
            scaling_table, list_of_dfs, scaling_mode, ressource_combinations = (
                self.__compute_scaling_metrics(region)
            )
            execution_modes = [df["executionMode"].to_numpy() for df in list_of_dfs]
            latest_date = None
            latest_change = None
            for df in list_of_dfs:
                date = pd.to_datetime(df["timestamp"].to_numpy().squeeze())
                git_commit = str(df["gitCommitShort"].to_numpy().squeeze())

                try:
                    if date > latest_date:
                        latest_date = date
                        latest_change = (date, git_commit)
                except TypeError:
                    latest_change = (date, git_commit)

            results.append(
                ScalingAnalysisResult(
                    region=region,
                    scaling_mode=scaling_mode,
                    scaling_table=scaling_table,
                    execution_modes=execution_modes,
                    latest_change=latest_change,
                    ressource_combinations=ressource_combinations,
                    dataframe_list=list_of_dfs,
                )
            )

        return results



from talp_pages.html.site import HTMLChildPage
from talp_pages.analysis.scaling_analysis import ScalingAnalysisResult
from typing import List
from talp_pages.plots.scaling_table_html_plot import ScalingTableHtmlPlot

from operator import itemgetter
import logging


class ScalingResultSite(HTMLChildPage):
    def __init__(self, results: List[ScalingAnalysisResult]) -> None:
        super().__init__(
            template_name="scaling.jinja", file_name="scaling.html", icon="bi-graph-up"
        )
        self.results = results
        self.latest_change = max(
            [result.latest_change for result in self.results], key=itemgetter(1)
        )  # Get newest change

    def get_content(self):
        if not self.results:
            raise ValueError("Handed no results to ScalingResultSite")

        regions = [result.region for result in self.results]
        scaling_mode = self.results[0].scaling_mode  # just take the first

        configurations = {}
        regions_to_render = []

        for region_name, result in zip(regions, self.results):
            configurations[region_name] = []
            plot = ScalingTableHtmlPlot(result)
            try:
                figure = plot.get_html()
            except ValueError as e:
                logging.warning(
                    f"Trying to get region with name  {region_name} for scaling analysis but failed. Skipping that region... {e}"
                )

            for datapoint in result.dataframe_list:
                configurations[region_name].append(
                    {
                        "label": datapoint["ressourceLabel"].to_numpy().squeeze(),
                        "json_file": datapoint["jsonFile"].to_numpy().squeeze(),
                        "timestamp": datapoint["timestamp"].to_numpy().squeeze(),
                        "metadata": datapoint["metadata"].to_numpy().squeeze(),
                    }
                )
                scaling_mode = datapoint["scalingMode"].to_numpy().squeeze()

            if figure:
                regions_to_render.append({"name": region_name, "figure": figure})

        html = self.render_template(
            regions=regions_to_render,
            configurations=configurations,
            scaling_mode=scaling_mode,
        )
        return html



from dataclasses import dataclass
from pathlib import Path
from typing import List
import os
from datetime import datetime
from talp_pages.common import TALP_JSON_TIMESTAMP_KEY
import logging
import json


@dataclass
class RunFolder:
    """Class for representing a Folder with a bunch of JSON files that are semantically connected through the folder strucutre"""

    relative_path: Path
    jsons: List[Path]
    latest_json: Path


def __set_newest_json(run_folder: RunFolder):
    newest_json = None
    newest_json_timestamp = None

    for json_entry in run_folder.jsons:
        with open(json_entry) as file:
            logging.debug(f"Opening {json_entry}")
            run_json = json.load(file)
            try:
                timestamp = datetime.fromisoformat(run_json[TALP_JSON_TIMESTAMP_KEY])
            except Exception:
                logging.critical(
                    f"Reading the {TALP_JSON_TIMESTAMP_KEY} key failed in {json_entry}. Falling back to the OS file creating date"
                )
                sys_time = os.path.getmtime(json_entry)
                timestamp = datetime.fromtimestamp(sys_time)

            if not newest_json_timestamp:  # base case
                newest_json_timestamp = timestamp
                newest_json = json_entry

            if timestamp.replace(tzinfo=None) > newest_json_timestamp.replace(
                tzinfo=None
            ):
                newest_json = json_entry
                newest_json_timestamp = timestamp

    run_folder.latest_json = newest_json


def get_run_folders(base_path: Path) -> List[RunFolder]:
    """Searches the base path and constructs the RunFolders it encouters, where at least one json file is contained"""
    run_folders = []

    for root, _, files in os.walk(base_path):
        json_files = [Path(root) / file for file in files if file.endswith(".json")]
        if json_files:
            relative_root = Path(root).relative_to(base_path)
            # dont add the most top level json
            if str(relative_root) != ".":
                run_folder = RunFolder(
                    relative_path=relative_root, jsons=json_files, latest_json=None
                )
                __set_newest_json(run_folder)
                run_folders.append(run_folder)
    return run_folders



from talp_pages.io.run_folders import RunFolder
from talp_pages.io.dataframe_handling import get_dfs
from dataclasses import dataclass
from typing import List

import logging


@dataclass
class BadgeResult:
    parallel_eff: float
    ressource_configuration: str


class BadgeAnalysis:
    def __init__(self, run_folder: RunFolder, region_for_badge: str) -> None:
        if not run_folder or not region_for_badge:
            raise ValueError("Badge Analyis cannot be constructed")
        self.run_folder = run_folder
        # get df from run_folders
        self.region = region_for_badge
        self.df = get_dfs([self.run_folder], as_list=False)

    def get_results(self, ressource_configurations: List[str]) -> List[BadgeResult]:
        results = []
        for rc in ressource_configurations:
            ressource_label_matching = self.df["ressourceLabel"] == rc
            region_matching = self.df["regionName"] == self.region

            target_df = self.df.loc[ressource_label_matching & region_matching]
            if not target_df.empty:
                parallel_efficiency = (
                    target_df.sort_values(by=["timestamp"], ascending=False)
                    .loc[:, "parallelEfficiency"]
                    .head(1)
                    .values[0]
                )
                results.append(
                    BadgeResult(
                        parallel_eff=parallel_efficiency, ressource_configuration=rc
                    )
                )
            else:
                logging.error(
                    f"Unable to find region {self.region} in {rc} for {self.df['runFolder'].head(1)}. Continuing with erroneous badge."
                )
                results.append(BadgeResult(parallel_eff=0, ressource_configuration=rc))

        return results



#!/usr/bin/env python

import os
import argparse
import gitlab
import logging

"""
    Script to automagically choose a suitable authentication method and download the .zip archive of the last sucessfull 
    execution of that jobname on the branch: ref-name
    Requires python-gitlab to be installed

    Default behaviour: create empty zip file
"""


def add_arguments(parser: argparse.ArgumentParser):
    parser.description = "Download artifacts from the latest successful GitLab CI job"
    parser.add_argument(
        "--gitlab-url", help="GitLab URL (the thing before /api/*)", required=True
    )
    parser.add_argument(
        "--project-name",
        help="GitLab project name (format: namespace/project)",
        required=True,
    )
    parser.add_argument(
        "--job-name",
        help="Name of the GitLab CI job from which to download the artifact",
        required=True,
    )
    parser.add_argument(
        "--output-file", help="Output file name for downloaded artifacts", required=True
    )
    parser.add_argument("--ref-name", help="GitLab branch name", default="main")
    parser.add_argument(
        "--gitlab-token",
        help="Personal GitLab access token (if not specified it will try to use a job_token)",
    )
    parser.add_argument(
        "--log-level",
        help="Logging level (DEBUG, INFO, WARNING, ERROR)",
        default="INFO",
    )


def download_artifacts(
    gl, project_name, job_name, ref_name, gitlab_token=None, output_file=None
):
    try:
        project = gl.projects.get(project_name)
    except gitlab.exceptions.GitlabGetError as e:
        logging.error("Failed to get project details: %s", str(e))
        return

    logging.info("Found project ID: %s", project.id)

    # Get artifacts for the job
    try:
        artifacts_file = project.artifacts.download(
            ref_name=ref_name, job=job_name.strip()
        )
        logging.info("Artifacts downloaded successfully: %s", output_file)
    except Exception:
        logging.critical(
            "Wasnt able to download artifacts, continue with a empty zip directory"
        )
        # The binary string is just the bare minimum to have a valid empty zip file
        artifacts_file = b"PK\x05\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"

    with open(output_file, "wb") as f:
        f.write(artifacts_file)


def main(args):
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        raise ValueError("Invalid log level: %s" % args.log_level)

    logging.basicConfig(level=log_level)
    job_gitlab_token = None
    gitlab_token = None

    if args.gitlab_token:
        gitlab_token = args.gitlab_token
    else:
        job_gitlab_token = os.getenv("CI_JOB_TOKEN")
        logging.debug("Using the CI_JOB_TOKEN variant")

    gl = None
    if gitlab_token:
        gl = gitlab.Gitlab(args.gitlab_url, private_token=gitlab_token)
        gl.auth()
    elif job_gitlab_token:
        gl = gitlab.Gitlab(args.gitlab_url, job_token=job_gitlab_token)
        logging.debug("Using the CI_JOB_TOKEN variant")
    else:
        gl = gitlab.Gitlab(args.gitlab_url)

    download_artifacts(
        gl,
        args.project_name,
        args.job_name,
        args.ref_name,
        args.gitlab_token,
        args.output_file,
    )



#!/usr/bin/env python

import argparse
import logging
import requests

"""
    Script to automagically download the .zip archive of the last sucessfull 
    execution of that jobname on the branch: ref-name
    Requires python-gitlab to be installed

    Default behaviour: create empty zip file
"""


def add_arguments(parser: argparse.ArgumentParser):
    parser.description = (
        "Download artifacts from the latest successful GitHub job in that workflow"
    )
    parser.add_argument(
        "--project-name",
        help="Github project name (format: namespace/project)",
        required=True,
    )
    parser.add_argument(
        "--workflow-name",
        help="Name of the workflow to download the artifacts from",
        required=True,
    )
    parser.add_argument(
        "--artifact-name",
        help="Name of the artifact to download",
        required=True,
    )
    parser.add_argument(
        "--output-file", help="Output file name for downloaded artifacts", required=True
    )
    parser.add_argument("--ref-name", help="Github branch name", default="main")
    parser.add_argument(
        "--github-token",
        help="Personal GitHub access token with enough permissions",
    )
    parser.add_argument(
        "--log-level",
        help="Logging level (DEBUG, INFO, WARNING, ERROR)",
        default="INFO",
    )


def download_artifacts(headers, download_url, output_file=None):
    # Get artifacts for the job
    try:
        download_resp = requests.get(download_url, headers=headers)
        download_resp.raise_for_status()
        artifacts_file = download_resp.content
        logging.info("Artifacts downloaded successfully: %s", output_file)
    except Exception:
        logging.critical(
            "Was not able to download artifacts, continue with a empty zip directory"
        )
        # The binary string is just the bare minimum to have a valid empty zip file
        artifacts_file = b"PK\x05\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    with open(output_file, "wb") as f:
        f.write(artifacts_file)


def get_workflow_id(headers, namespace, workflow_name):
    workflow_url = f"https://api.github.com/repos/{namespace}/actions/workflows"
    workflow_resp = requests.get(workflow_url, headers=headers)
    workflow_resp.raise_for_status()
    workflows = workflow_resp.json()["workflows"]
    workflow_id = next(
        (
            wf["id"]
            for wf in workflows
            if wf["name"] == workflow_name or wf["path"].endswith(workflow_name)
        ),
        None,
    )
    if not workflow_id:
        raise Exception(f'Workflow "{workflow_name}" not found.')

    return workflow_id


def get_latest_run_id(headers, namespace, workflow_id):
    runs_url = (
        f"https://api.github.com/repos/{namespace}/actions/workflows/{workflow_id}/runs"
    )
    runs_resp = requests.get(
        runs_url, headers=headers, params={"status": "completed", "per_page": 1}
    )
    runs_resp.raise_for_status()
    runs = runs_resp.json()["workflow_runs"]
    if not runs:
        raise ValueError("No sucessfully completed workflow runs found.")
    return runs[0]["id"]


def get_artifact_url(headers, namespace, latest_run_id, artifact_name):
    artifacts_url = f"https://api.github.com/repos/{namespace}/actions/runs/{latest_run_id}/artifacts"
    artifacts_resp = requests.get(artifacts_url, headers=headers)
    artifacts_resp.raise_for_status()
    artifacts = artifacts_resp.json()["artifacts"]

    artifact = next((a for a in artifacts if a["name"] == artifact_name), None)
    if not artifact:
        raise ValueError(
            f'Artifact "{artifact_name}" not found in workflow run with id: {latest_run_id}'
        )
    return artifact["archive_download_url"]


def main(args):
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        raise ValueError("Invalid log level: %s" % args.log_level)

    logging.basicConfig(level=log_level)

    # Get token
    http_request_headers = {
        "Authorization": f"token {args.github_token}",
        "Accept": "application/vnd.github+json",
    }
    # TODO add some sanity checks on project_name

    workflow_id = get_workflow_id(
        headers=http_request_headers,
        namespace=args.project_name,
        workflow_name=args.workflow_name,
    )
    try:
        latest_run_id = get_latest_run_id(
            headers=http_request_headers,
            workflow_id=workflow_id,
            namespace=args.project_name,
        )
    except ValueError as e:
        logging.critical("Unable to obtain the last run.")
        logging.critical(f"{e}")
        latest_run_id = ""
    try:
        download_url = get_artifact_url(
            headers=http_request_headers,
            namespace=args.project_name,
            latest_run_id=latest_run_id,
            artifact_name=args.artifact_name,
        )
    except ValueError as e:
        logging.critical("Could not get the download URL of the artifact.")
        logging.critical(f"{e}")
        download_url = ""

    download_artifacts(
        headers=http_request_headers,
        download_url=download_url,
        output_file=args.output_file,
    )



from typing import List
from pathlib import Path
from dataclasses import dataclass
from operator import itemgetter
from tqdm import tqdm

from talp_pages.io.run_folders import RunFolder
from talp_pages.analysis.scaling_analysis import (
    ScalingAnalysis,
    ScalingAnalysisArguments,
)
from talp_pages.analysis.time_series_analysis import (
    TimeSeriesAnalysis,
    TimeSeriesAnalysisArguments,
)
from talp_pages.analysis.badge_analysis import BadgeAnalysis
from talp_pages.io.dataframe_handling import get_ressource_combinations
from talp_pages.html.scaling_result_site import ScalingResultSite
from talp_pages.html.time_series_result_site import TimeSeriesSite
from talp_pages.html.ci_report_index_page import CIReportIndexPage
from talp_pages.svg.badge import Badge
from talp_pages.io.output import write_badge, write_site


@dataclass
class CIReportArgs:
    run_folders: List[RunFolder]
    selected_regions: List[str]
    region_for_badge: str
    with_badge: bool


class CIReport:
    def __init__(self, args: CIReportArgs) -> None:
        self.run_folders = args.run_folders
        self.selected_regions = args.selected_regions
        self.region_for_badge = args.region_for_badge
        self.with_badge = args.with_badge

    def write_to_folder(self, output_path: Path):
        sites = {}
        # for each run folder get the sites
        for run_folder in tqdm(self.run_folders, desc="Processing subfolders"):
            # first get scaling result site one per region
            scaling_results = ScalingAnalysis(
                ScalingAnalysisArguments(run_folder=run_folder)
            ).get_results(self.selected_regions)
            scaling_result_site = ScalingResultSite(scaling_results)
            # one entry per ressource combination
            ressource_configurations = get_ressource_combinations(run_folder)
            time_series_results = TimeSeriesAnalysis(
                TimeSeriesAnalysisArguments(run_folder=run_folder)
            ).get_results(ressource_configurations)

            time_series_sites = []
            for rc, result in zip(ressource_configurations, time_series_results):
                time_series_sites.append(TimeSeriesSite(result, self.selected_regions))

            badges = []
            if self.with_badge:
                bagde_results = BadgeAnalysis(
                    run_folder=run_folder, region_for_badge=self.region_for_badge
                ).get_results(ressource_configurations=ressource_configurations)
                for rc, result in zip(ressource_configurations, bagde_results):
                    badges.append(Badge(result))

            sites[run_folder.relative_path] = {
                "scaling_result_site": scaling_result_site,
                "ressource_configurations": ressource_configurations,
                "badges": badges,
                "region_for_badge": self.region_for_badge,
                "time_series_sites": time_series_sites,
                "relative_path": run_folder.relative_path,
                "latest_change": max(
                    [
                        site.latest_change
                        for site in [scaling_result_site, *time_series_sites]
                    ],
                    key=itemgetter(1),
                ),  # Get newest across subsites
            }
            # Write out the things
            subfolder = output_path.joinpath(run_folder.relative_path)
            for site in [scaling_result_site, *time_series_sites]:
                write_site(subfolder, site)
            for badge in badges:
                write_badge(subfolder, badge)

        index_page = CIReportIndexPage(subsites=sites)
        write_site(output_path, index_page)
